from __future__ import annotations
import os
import re
import logging
from typing_extensions import Annotated
from typing import (
    ClassVar,
    Collection,
    Final,
    List,
    Optional,
    TypeVar,
)


import toml
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, BaseSettings, validator, PrivateAttr, DirectoryPath
from pydantic.networks import HttpUrl, EmailStr

import typer

from ..bili.typing import QualityNumber
from ..postprocess import DeleteStrategy
from ..logging.typing import LOG_LEVEL
from ..utils.string import camel_case
from ..path.helpers import file_exists, create_file


logger = logging.getLogger(__name__)


__all__ = (
    'DEFAULT_SETTINGS_PATH',

    'EnvSettings',
    'Settings',
    'SettingsIn',
    'SettingsOut',

    'HeaderOptions',
    'HeaderSettings',
    'DanmakuOptions',
    'DanmakuSettings',
    'RecorderOptions',
    'RecorderSettings',
    'PostprocessingSettings',
    'PostprocessingOptions',

    'TaskOptions',
    'TaskSettings',
    'OutputSettings',
    'LoggingSettings',
    'SpaceSettings',
    'EmailSettings',
    'ServerchanSettings',
    'PushplusSettings',
    'NotifierSettings',
    'NotificationSettings',
    'EmailNotificationSettings',
    'ServerchanNotificationSettings',
    'PushplusNotificationSettings',
    'WebHookSettings',
)


DEFAULT_SETTINGS_PATH: Final[str] = '~/.blrec/settings.toml'


def settings_file_factory() -> str:
    path = os.path.abspath(os.path.expanduser(DEFAULT_SETTINGS_PATH))
    if not file_exists(path):
        create_file(path)
        typer.secho(
            f"Created setting file: '{path}'",
            fg=typer.colors.BRIGHT_MAGENTA,
            bold=True,
        )
    return path


class EnvSettings(BaseSettings):
    settings_file: Annotated[
        str, Field(env='config', default_factory=settings_file_factory)
    ]
    out_dir: Optional[str] = None
    api_key: Annotated[
        Optional[str],
        Field(min_length=8, max_length=80, regex=r'[a-zA-Z\d\-]{8,80}'),
    ] = None

    class Config:
        anystr_strip_whitespace = True


_V = TypeVar('_V')


class BaseModel(PydanticBaseModel):
    class Config:
        extra = 'forbid'
        validate_assignment = True
        anystr_strip_whitespace = True
        allow_population_by_field_name = True

        @classmethod
        def alias_generator(cls, string: str) -> str:
            return camel_case(string)

    @staticmethod
    def _validate_with_collection(
        value: _V, allowed_values: Collection[_V]
    ) -> None:
        if value not in allowed_values:
            raise ValueError(
                f'the value {value} does not be allowed, '
                f'must be one of {", ".join(map(str, allowed_values))}'
            )


class HeaderOptions(BaseModel):
    user_agent: Optional[str]
    cookie: Optional[str]


class HeaderSettings(HeaderOptions):
    user_agent: str = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) ' \
        'AppleWebKit/537.36 (KHTML, like Gecko) ' \
        'Chrome/89.0.4389.114 Safari/537.36'
    cookie: str = ''


class DanmakuOptions(BaseModel):
    danmu_uname: Optional[bool]


class DanmakuSettings(DanmakuOptions):
    danmu_uname: bool = False


class RecorderOptions(BaseModel):
    quality_number: Optional[QualityNumber]
    read_timeout: Optional[int]  # seconds
    buffer_size: Annotated[  # bytes
        Optional[int], Field(ge=4096, le=1024 ** 2 * 512, multiple_of=2)
    ]

    @validator('read_timeout')
    def _validate_read_timeout(cls, value: Optional[int]) -> Optional[int]:
        if value is not None:
            allowed_values = frozenset((3, 5, 10, 30, 60, 180, 300, 600))
            cls._validate_with_collection(value, allowed_values)
        return value


class RecorderSettings(RecorderOptions):
    quality_number: QualityNumber = 20000  # 4K, the highest quality.
    read_timeout: int = 3
    buffer_size: Annotated[
        int, Field(ge=4096, le=1024 ** 2 * 512, multiple_of=2)
    ] = 8192


class PostprocessingOptions(BaseModel):
    remux_to_mp4: Optional[bool]
    delete_source: Optional[DeleteStrategy]


class PostprocessingSettings(PostprocessingOptions):
    remux_to_mp4: bool = False
    delete_source: DeleteStrategy = DeleteStrategy.AUTO


class OutputOptions(BaseModel):
    path_template: Optional[str]
    filesize_limit: Optional[int]  # file size in bytes
    duration_limit: Optional[int]  # duration in seconds

    @validator('path_template')
    def _validate_path_template(cls, value: Optional[str]) -> Optional[str]:
        if value is not None:
            pattern = r'''^
                (?:
                    [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                    \{
                    (?:
                        roomid|uname|title|area|parent_area|
                        year|month|day|hour|minute|second
                    )
                    \}
                    [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                )+?
                (?:
                    /
                    (?:
                        [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                        \{
                        (?:
                            roomid|uname|title|area|parent_area|
                            year|month|day|hour|minute|second
                        )
                        \}
                        [^\\/:*?"<>|\t\n\r\f\v\{\}]*?
                    )+?
                )*
            $'''
            if not re.fullmatch(pattern, value, re.VERBOSE):
                raise ValueError(f"invalid path template: '{value}'")
        return value

    @validator('filesize_limit')
    def _validate_filesize_limit(cls, value: Optional[int]) -> Optional[int]:
        # allowed 1 ~ 20 GB, 0 indicates not limit.
        if value is not None:
            allowed_values = frozenset(1024 ** 3 * i for i in range(0, 21))
            cls._validate_with_collection(value, allowed_values)
        return value

    @validator('duration_limit')
    def _validate_duration_limit(cls, value: Optional[int]) -> Optional[int]:
        # allowed 1 ~ 24 hours, 0 indicates not limit.
        if value is not None:
            allowed_values = frozenset(3600 * i for i in range(0, 25))
            cls._validate_with_collection(value, allowed_values)
        return value


class OutputSettings(OutputOptions):
    out_dir: Annotated[str, DirectoryPath] = '.'
    path_template: str = (
        '{roomid} - {uname}/'
        'blive_{roomid}_{year}-{month}-{day}-{hour}{minute}{second}'
    )
    filesize_limit: int = 0  # no limit by default
    duration_limit: int = 0  # no limit by default

    @validator('out_dir')
    def _validate_dir(cls, path: str) -> str:
        if not os.path.isdir(os.path.expanduser(path)):
            raise ValueError(f"'{path}' not a directory")
        return path


class TaskOptions(BaseModel):
    output: OutputOptions = OutputOptions()
    header: HeaderOptions = HeaderOptions()
    danmaku: DanmakuOptions = DanmakuOptions()
    recorder: RecorderOptions = RecorderOptions()
    postprocessing: PostprocessingOptions = PostprocessingOptions()

    @classmethod
    def from_settings(cls, settings: TaskSettings) -> TaskOptions:
        return cls(**settings.dict(
            include={
                'output', 'header', 'danmaku', 'recorder', 'postprocessing'
            }
        ))


class TaskSettings(TaskOptions):
    # must use the real room id rather than the short room id!
    room_id: Annotated[int, Field(ge=1, le=99999999)]
    enable_monitor: bool = True
    enable_recorder: bool = True


class LoggingSettings(BaseModel):
    console_log_level: LOG_LEVEL = 'INFO'
    max_bytes: Annotated[
        int, Field(ge=1024 ** 2, le=1024 ** 2 * 10, multiple_of=1024 ** 2)
    ] = 1024 ** 2 * 10  # allowed 1 ~ 10 MB
    backup_count: Annotated[int, Field(ge=1, le=30)] = 30


class SpaceSettings(BaseModel):
    check_interval: int = 60  # 1 minutes
    space_threshold: int = 1024 ** 3  # 1 GB
    recycle_records: bool = False

    @validator('check_interval')
    def _validate_interval(cls, value: int) -> int:
        allowed_values = frozenset(60 * i for i in (1, 3, 5, 10))
        cls._validate_with_collection(value, allowed_values)
        return value

    @validator('space_threshold')
    def _validate_threshold(cls, value: int) -> int:
        allowed_values = frozenset(1024 ** 3 * i for i in (1, 3, 5, 10))
        cls._validate_with_collection(value, allowed_values)
        return value


class EmailSettings(BaseModel):
    src_addr: Annotated[str, EmailStr] = ''
    dst_addr: Annotated[str, EmailStr] = ''
    auth_code: str = ''
    smtp_host: str = 'smtp.163.com'
    smtp_port: int = 465


class ServerchanSettings(BaseModel):
    sendkey: str = ''

    @validator('sendkey')
    def _validate_sendkey(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'[a-zA-Z\d]+', value):
            raise ValueError('sendkey is invalid')
        return value


class PushplusSettings(BaseModel):
    token: str = ''
    topic: str = ''

    @validator('token')
    def _validate_token(cls, value: str) -> str:
        if value != '' and not re.fullmatch(r'[a-z\d]{32}', value):
            raise ValueError('token is invalid')
        return value


class NotifierSettings(BaseModel):
    enabled: bool = False


class NotificationSettings(BaseModel):
    notify_began: bool = True
    notify_ended: bool = True
    notify_error: bool = True
    notify_space: bool = True


class EmailNotificationSettings(
    EmailSettings, NotifierSettings, NotificationSettings
):
    pass


class ServerchanNotificationSettings(
    ServerchanSettings, NotifierSettings, NotificationSettings
):
    pass


class PushplusNotificationSettings(
    PushplusSettings, NotifierSettings, NotificationSettings
):
    pass


class WebHookEventSettings(BaseModel):
    live_began: bool = True
    live_ended: bool = True
    room_change: bool = True
    space_no_enough: bool = True
    file_completed: bool = True
    error_occurred: bool = True


class WebHookSettings(WebHookEventSettings):
    url: Annotated[str, HttpUrl]


class Settings(BaseModel):
    _MAX_TASKS: ClassVar[int] = 100
    _MAX_WEBHOOKS: ClassVar[int] = 50

    _path: str = PrivateAttr()
    version: str = '1.0'

    tasks: Annotated[List[TaskSettings], Field(max_items=100)] = []
    output: OutputSettings = OutputSettings()
    logging: LoggingSettings = LoggingSettings()
    header: HeaderSettings = HeaderSettings()
    danmaku: DanmakuSettings = DanmakuSettings()
    recorder: RecorderSettings = RecorderSettings()
    postprocessing: PostprocessingSettings = PostprocessingSettings()
    space: SpaceSettings = SpaceSettings()
    email_notification: EmailNotificationSettings = EmailNotificationSettings()
    serverchan_notification: ServerchanNotificationSettings = \
        ServerchanNotificationSettings()
    pushplus_notification: PushplusNotificationSettings = \
        PushplusNotificationSettings()
    webhooks: Annotated[List[WebHookSettings], Field(max_items=50)] = []

    @classmethod
    def load(cls, path: str) -> Settings:
        settings = cls.parse_obj(toml.load(path))
        settings._path = path
        return settings

    def update_from_env_settings(self, env_settings: EnvSettings) -> None:
        if (out_dir := env_settings.out_dir) is not None:
            self.output.out_dir = out_dir

    def dump(self) -> None:
        assert self._path
        with open(self._path, 'wt', encoding='utf8') as file:
            toml.dump(self.dict(exclude_none=True), file)

    @validator('tasks')
    def _validate_tasks(cls, tasks: List[TaskSettings]) -> List[TaskSettings]:
        if len(tasks) >= cls._MAX_TASKS:
            raise ValueError(f'Out of max tasks limits: {cls._MAX_TASKS}')
        return tasks

    @validator('webhooks')
    def _validate_webhooks(
        cls, webhooks: List[WebHookSettings]
    ) -> List[WebHookSettings]:
        if len(webhooks) >= cls._MAX_WEBHOOKS:
            raise ValueError(
                f'Out of max webhooks limits: {cls._MAX_WEBHOOKS}'
            )
        return webhooks


class SettingsIn(BaseModel):
    output: Optional[OutputSettings] = None
    logging: Optional[LoggingSettings] = None
    header: Optional[HeaderSettings] = None
    danmaku: Optional[DanmakuSettings] = None
    recorder: Optional[RecorderSettings] = None
    postprocessing: Optional[PostprocessingSettings] = None
    space: Optional[SpaceSettings] = None
    email_notification: Optional[EmailNotificationSettings] = None
    serverchan_notification: Optional[ServerchanNotificationSettings] = None
    pushplus_notification: Optional[PushplusNotificationSettings] = None
    webhooks: Optional[List[WebHookSettings]] = None


class SettingsOut(SettingsIn):
    version: Optional[str] = None
    tasks: Optional[List[TaskSettings]] = None



from typing import cast

import attr

from .struct_io import StructReader, StructWriter
from .io_protocols import RandomIO
from .models import (
    FlvTag,
    TagType,
    FlvHeader,
    FlvTagHeader,

    AudioTag,
    AudioTagHeader,
    SoundFormat,
    SoundRate,
    SoundSize,
    SoundType,
    AACPacketType,

    VideoTag,
    VideoTagHeader,
    CodecID,
    FrameType,
    AVCPacketType,

    ScriptTag,
)


__all__ = 'FlvParser', 'FlvDumper'


class FlvParser:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream
        self._reader = StructReader(stream)

    def parse_header(self) -> FlvHeader:
        signature = self._reader.read(3).decode()
        assert signature == 'FLV', 'Not a FLV file'
        version = self._reader.read_ui8()
        type_flag = self._reader.read_ui8()
        data_offset = self._reader.read_ui32()
        return FlvHeader(signature, version, type_flag, data_offset)

    def parse_previous_tag_size(self) -> int:
        return self._reader.read_ui32()

    def parse_tag(self, *, no_body: bool = False) -> FlvTag:
        offset = self._stream.tell()
        tag_header = self.parse_flv_tag_header()

        tag: FlvTag
        arguments = dict(attr.asdict(tag_header), offset=offset)

        if tag_header.tag_type == TagType.AUDIO:
            audio_tag_header = self.parse_audio_tag_header()
            arguments.update(attr.asdict(audio_tag_header))
            tag = AudioTag(**arguments)
        elif tag_header.tag_type == TagType.VIDEO:
            video_tag_header = self.parse_video_tag_header()
            arguments.update(attr.asdict(video_tag_header))
            tag = VideoTag(**arguments)
        elif tag_header.tag_type == TagType.SCRIPT:
            tag = ScriptTag(**arguments)
        else:
            raise ValueError(tag_header.tag_type)

        if no_body:
            self._stream.seek(tag.tag_end_offset)
        else:
            body = self._reader.read(tag.body_size)
            tag = tag.evolve(body=body)

        return tag

    def parse_flv_tag_header(self) -> FlvTagHeader:
        flag = self._reader.read_ui8()
        filtered = bool(flag & 0b0010_0000)
        assert not filtered, 'Unsupported Filtered FLV Tag'
        tag_type = TagType(flag & 0b0001_1111)
        data_size = self._reader.read_ui24()
        assert data_size > 0, 'Invalid Tag'
        timestamp = self._reader.read_ui24()
        timestamp_extended = self._reader.read_ui8()
        timestamp = timestamp_extended << 24 | timestamp
        stream_id = self._reader.read_ui24()
        return FlvTagHeader(
            filtered, tag_type, data_size, timestamp, stream_id
        )

    def parse_audio_tag_header(self) -> AudioTagHeader:
        flag = self._reader.read_ui8()
        sound_format = SoundFormat(flag >> 4)
        sound_rate = SoundRate((flag >> 2) & 0b0000_0011)
        sound_size = SoundSize((flag >> 1) & 0b0000_0001)
        sound_type = SoundType(flag & 0b0000_0001)

        if sound_format != SoundFormat.AAC:
            aac_packet_type = None
        else:
            aac_packet_type = AACPacketType(self._reader.read_ui8())

        return AudioTagHeader(
            sound_format, sound_rate, sound_size, sound_type, aac_packet_type
        )

    def parse_video_tag_header(self) -> VideoTagHeader:
        flag = self._reader.read_ui8()
        frame_type = FrameType(flag >> 4)
        codec_id = CodecID(flag & 0b0000_1111)

        if codec_id != CodecID.AVC:
            avc_packet_type = None
            composition_time = None
        else:
            avc_packet_type = AVCPacketType(self._reader.read_ui8())
            composition_time = self._reader.read_ui24()

        return VideoTagHeader(
            frame_type, codec_id, avc_packet_type, composition_time
        )


class FlvDumper:
    def __init__(self, stream: RandomIO) -> None:
        self._stream = stream
        self._writer = StructWriter(stream)

    def dump_header(self, flv_header: FlvHeader) -> None:
        self._writer.write(flv_header.signature.encode())
        self._writer.write_ui8(flv_header.version)
        self._writer.write_ui8(flv_header.type_flag)
        self._writer.write_ui32(flv_header.data_offset)

    def dump_previous_tag_size(self, size: int) -> None:
        self._writer.write_ui32(size)

    def dump_tag(self, tag: FlvTag) -> None:
        self.dump_flv_tag_header(tag)

        if tag.is_audio_tag():
            audio_tag = cast(AudioTag, tag)
            self.dump_audio_tag_header(audio_tag)
        elif tag.is_video_tag():
            video_tag = cast(VideoTag, tag)
            self.dump_video_tag_header(video_tag)
        elif tag.is_script_tag():
            pass
        else:
            raise ValueError(tag.tag_type)

        if tag.body is None:
            self._stream.seek(tag.tag_end_offset)
        else:
            self._writer.write(tag.body)

    def dump_flv_tag_header(self, tag: FlvTag) -> None:
        self._writer.write_ui8((int(tag.filtered) << 5) | tag.tag_type.value)
        self._writer.write_ui24(tag.data_size)
        self._writer.write_ui24(tag.timestamp & 0x00ffffff)
        self._writer.write_ui8(tag.timestamp >> 24)
        self._writer.write_ui24(tag.stream_id)

    def dump_audio_tag_header(self, tag: AudioTag) -> None:
        self._writer.write_ui8(
            (tag.sound_format.value << 4) |
            (tag.sound_rate.value << 2) |
            (tag.sound_size.value << 1) |
            tag.sound_type.value
        )
        if tag.aac_packet_type is not None:
            assert tag.sound_format == SoundFormat.AAC
            self._writer.write_ui8(tag.aac_packet_type)

    def dump_video_tag_header(self, tag: VideoTag) -> None:
        self._writer.write_ui8(
            (tag.frame_type.value << 4) | tag.codec_id.value
        )
        if tag.codec_id == CodecID.AVC:
            assert tag.avc_packet_type is not None
            assert tag.composition_time is not None
            self._writer.write_ui8(tag.avc_packet_type.value)
            self._writer.write_ui24(tag.composition_time)



import os
import logging
from typing import Any, Dict, Optional

import attr
from rx import create, operators as op
from rx.subject import Subject
from rx.core import Observable
from rx.core.typing import Observer, Scheduler, Disposable
from rx.scheduler.currentthreadscheduler import CurrentThreadScheduler
from tqdm import tqdm

from .stream_processor import StreamProcessor, BaseOutputFileManager, JoinPoint
from .helpers import get_metadata, make_comment_for_joinpoints
from ..logging.room_id import with_room_id


__all__ = 'MetadataInjector', 'InjectProgress', 'inject_metadata'


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class InjectProgress:
    time: int
    duration: int


class MetadataInjector:
    def __init__(self, in_path: str, out_path: str) -> None:
        self._in_path = in_path
        self._file_manager = OutputFileManager(out_path)
        self._duration: int = 0
        self._progress_updates = Subject()

    @property
    def progress_updates(self) -> Observable:
        return self._progress_updates

    def inject(self, in_metadata: Dict[str, Any]) -> None:
        metadata = get_metadata(self._in_path)
        metadata.update(in_metadata)

        self._duration = int(float(metadata['duration']) * 1000)
        self._progress_updates.on_next(InjectProgress(0, self._duration))

        self._append_comment_for_joinpoints(metadata)

        processor = StreamProcessor(
            self._file_manager,
            metadata=metadata,
            disable_limit=True,
        )

        def update_progress(time: int) -> None:
            progress = InjectProgress(time, self._duration)
            self._progress_updates.on_next(progress)

        processor.time_updates.subscribe(update_progress)

        with open(self._in_path, 'rb') as in_file:
            processor.process_stream(in_file)
            processor.finalize()

        progress = InjectProgress(self._duration, self._duration)
        self._progress_updates.on_next(progress)

    @staticmethod
    def _append_comment_for_joinpoints(metadata: Dict[str, Any]) -> None:
        if (join_points := metadata.get('joinpoints')):
            join_points = map(JoinPoint.from_metadata_value, join_points)
            if 'Comment' in metadata:
                metadata['Comment'] += '\n\n' + \
                    make_comment_for_joinpoints(join_points)
            else:
                metadata['Comment'] = make_comment_for_joinpoints(join_points)


class OutputFileManager(BaseOutputFileManager):
    def __init__(self, out_path: str) -> None:
        super().__init__()
        self._out_path = out_path

    def _make_path(self) -> str:
        return self._out_path


def inject_metadata(
    path: str,
    metadata: Dict[str, Any],
    *,
    report_progress: bool = False,
    room_id: Optional[int] = None,
) -> Observable:
    def subscribe(
        observer: Observer[InjectProgress],
        scheduler: Optional[Scheduler] = None,
    ) -> Disposable:
        _scheduler = scheduler or CurrentThreadScheduler.singleton()

        def action(scheduler, state):  # type: ignore
            root, ext = os.path.splitext(path)
            out_path = f'{root}_inject_metadata{ext}'
            injector = MetadataInjector(path, out_path)
            file_name = os.path.basename(path)

            with tqdm(desc='Injecting', unit='ms', postfix=file_name) as pbar:
                def reset(progress: InjectProgress) -> None:
                    pbar.reset(progress.duration)

                def update(progress: InjectProgress) -> None:
                    pbar.update(progress.time - pbar.n)

                injector.progress_updates.pipe(op.first()).subscribe(reset)
                injector.progress_updates.pipe(op.skip(1)).subscribe(update)

                if report_progress:
                    injector.progress_updates.subscribe(
                        lambda p: observer.on_next(p)
                    )

                try:
                    logger.info(f"Injecting metadata for '{path}' ...")
                    injector.inject(metadata)
                except Exception as e:
                    logger.error(
                        f"Failed to inject metadata for '{path}': {repr(e)}"
                    )
                    observer.on_error(e)
                else:
                    logger.info(f"Successfully inject metadata for '{path}'")
                    os.replace(out_path, path)
                    observer.on_completed()

        if room_id is not None:
            return _scheduler.schedule(with_room_id(room_id)(action))
        else:
            return _scheduler.schedule(action)

    return create(subscribe)



from __future__ import annotations
from abc import ABC
import uuid
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Generic, TYPE_CHECKING, Type, TypeVar

import attr

if TYPE_CHECKING:
    from ..bili.models import UserInfo, RoomInfo
    from ..disk_space.space_monitor import DiskUsage


_D = TypeVar('_D', bound='BaseEventData')
_E = TypeVar('_E')
# XXX error: Invalid self argument ...
# _E = TypeVar('_E', bound='BaseEvent[BaseEventData]')


class BaseEventData(ABC):
    pass


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class BaseEvent(Generic[_D]):
    type: str
    id: uuid.UUID
    date: datetime
    data: _D

    @classmethod
    def from_data(cls: Type[_E], data: _D) -> _E:
        return cls(  # type: ignore
            id=uuid.uuid1(),
            date=datetime.now(timezone(timedelta(hours=8))),
            data=data,
        )

    def asdict(self) -> Dict[str, Any]:
        return attr.asdict(self, value_serializer=value_serializer)


def value_serializer(instance, attribute, value):  # type: ignore
    if isinstance(value, (uuid.UUID, datetime)):
        return str(value)
    return value


@attr.s(auto_attribs=True, slots=True, frozen=True)
class LiveBeganEventData(BaseEventData):
    user_info: UserInfo
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class LiveBeganEvent(BaseEvent[LiveBeganEventData]):
    type: str = 'LiveBeganEvent'
    data: LiveBeganEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class LiveEndedEventData(LiveBeganEventData):
    pass


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class LiveEndedEvent(BaseEvent[LiveEndedEventData]):
    type: str = 'LiveEndedEvent'
    data: LiveEndedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class RoomChangeEventData(BaseEventData):
    room_info: RoomInfo


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class RoomChangeEvent(BaseEvent[RoomChangeEventData]):
    type: str = 'RoomChangeEvent'
    data: RoomChangeEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SpaceNoEnoughEventData(BaseEventData):
    path: str
    threshold: int
    usage: DiskUsage


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class SpaceNoEnoughEvent(BaseEvent[SpaceNoEnoughEventData]):
    type: str = 'SpaceNoEnoughEvent'
    data: SpaceNoEnoughEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class FileCompletedEventData(BaseEventData):
    room_id: int
    path: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class FileCompletedEvent(BaseEvent[FileCompletedEventData]):
    type: str = 'FileCompletedEvent'
    data: FileCompletedEventData


@attr.s(auto_attribs=True, slots=True, frozen=True)
class ErrorData(BaseEventData):
    name: str
    detail: str


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class Error(BaseEvent[ErrorData]):
    type: str = 'Error'
    data: ErrorData



import logging
from typing import List, Optional


from .models import RunningStatus, TaskStatus
from ..bili.live import Live
from ..bili.models import RoomInfo, UserInfo
from ..bili.danmaku_client import DanmakuClient
from ..bili.live_monitor import LiveMonitor
from ..bili.typing import QualityNumber
from ..core import Recorder
from ..postprocess import Postprocessor, ProcessStatus, DeleteStrategy
from ..event.event_submitters import (
    LiveEventSubmitter, PostprocessorEventSubmitter
)
from ..logging.room_id import aio_task_with_room_id


__all__ = 'RecordTask',


logger = logging.getLogger(__name__)


class RecordTask:
    def __init__(
        self,
        room_id: int,
        *,
        out_dir: str = '',
        path_template: str = '',
        cookie: str = '',
        user_agent: str = '',
        danmu_uname: bool = False,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
        remux_to_mp4: bool = False,
        delete_source: DeleteStrategy = DeleteStrategy.AUTO,
    ) -> None:
        super().__init__()

        self._live = Live(room_id, user_agent, cookie)

        self._room_id = room_id
        self._out_dir = out_dir
        self._path_template = path_template
        self._cookie = cookie
        self._user_agent = user_agent
        self._danmu_uname = danmu_uname
        self._buffer_size = buffer_size
        self._read_timeout = read_timeout
        self._filesize_limit = filesize_limit
        self._duration_limit = duration_limit
        self._remux_to_mp4 = remux_to_mp4
        self._delete_source = delete_source

        self._ready = False
        self._monitor_enabled: bool = False
        self._recorder_enabled: bool = False

    @property
    def ready(self) -> bool:
        return self._ready

    @property
    def room_info(self) -> RoomInfo:
        return self._live.room_info

    @property
    def user_info(self) -> UserInfo:
        return self._live.user_info

    @property
    def running_status(self) -> RunningStatus:
        if not self._monitor_enabled and not self._recorder_enabled:
            return RunningStatus.STOPPED
        elif self._recorder.recording:
            return RunningStatus.RECORDING
        elif self._postprocessor.status == ProcessStatus.REMUXING:
            return RunningStatus.REMUXING
        elif self._postprocessor.status == ProcessStatus.INJECTING:
            return RunningStatus.INJECTING
        else:
            return RunningStatus.WAITING

    @property
    def monitor_enabled(self) -> bool:
        return self._monitor_enabled

    @property
    def recorder_enabled(self) -> bool:
        return self._recorder_enabled

    @property
    def status(self) -> TaskStatus:
        return TaskStatus(
            monitor_enabled=self.monitor_enabled,
            recorder_enabled=self.recorder_enabled,
            running_status=self.running_status,
            elapsed=self._recorder.elapsed,
            data_count=self._recorder.data_count,
            data_rate=self._recorder.data_rate,
            danmu_count=self._recorder.danmu_count,
            danmu_rate=self._recorder.danmu_rate,
            real_quality_number=self._recorder.real_quality_number,
            postprocessing_path=self._postprocessor.postprocessing_path,
            postprocessing_progress=(
                self._postprocessor.postprocessing_progress
            ),
        )

    @property
    def files(self) -> List[str]:
        if self.running_status == RunningStatus.STOPPED:
            return []
        if (
            self.running_status == RunningStatus.RECORDING or
            not self._postprocessor.enabled
        ):
            return [
                *self._recorder.get_video_files(),
                *self._recorder.get_danmaku_files(),
            ]
        return [*self._postprocessor.get_final_files()]

    @property
    def user_agent(self) -> str:
        return self._live.user_agent

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self._live.user_agent = value

    @property
    def cookie(self) -> str:
        return self._live.cookie

    @cookie.setter
    def cookie(self, value: str) -> None:
        self._live.cookie = value

    @property
    def danmu_uname(self) -> bool:
        return self._recorder.danmu_uname

    @danmu_uname.setter
    def danmu_uname(self, value: bool) -> None:
        self._recorder.danmu_uname = value

    @property
    def quality_number(self) -> QualityNumber:
        return self._recorder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._recorder.quality_number = value

    @property
    def real_quality_number(self) -> QualityNumber:
        return self._recorder.real_quality_number

    @property
    def buffer_size(self) -> int:
        return self._recorder.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._recorder.buffer_size = value

    @property
    def read_timeout(self) -> int:
        return self._recorder.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._recorder.read_timeout = value

    @property
    def out_dir(self) -> str:
        return self._recorder.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._recorder.out_dir = value

    @property
    def path_template(self) -> str:
        return self._recorder.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._recorder.path_template = value

    @property
    def filesize_limit(self) -> int:
        return self._recorder.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._recorder.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._recorder.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._recorder.duration_limit = value

    @property
    def remux_to_mp4(self) -> bool:
        return self._postprocessor.remux_to_mp4

    @remux_to_mp4.setter
    def remux_to_mp4(self, value: bool) -> None:
        self._postprocessor.remux_to_mp4 = value

    @property
    def delete_source(self) -> DeleteStrategy:
        return self._postprocessor.delete_source

    @delete_source.setter
    def delete_source(self, value: DeleteStrategy) -> None:
        self._postprocessor.delete_source = value

    @aio_task_with_room_id
    async def setup(self) -> None:
        await self._live.init()
        await self._setup()
        self._ready = True

    @aio_task_with_room_id
    async def destroy(self) -> None:
        await self._destroy()
        await self._live.deinit()
        self._ready = False

    @aio_task_with_room_id
    async def enable_monitor(self) -> None:
        if self._monitor_enabled:
            return
        self._monitor_enabled = True

        await self._danmaku_client.start()
        self._live_monitor.enable()

    @aio_task_with_room_id
    async def disable_monitor(self) -> None:
        if not self._monitor_enabled:
            return
        self._monitor_enabled = False

        self._live_monitor.disable()
        await self._danmaku_client.stop()

    @aio_task_with_room_id
    async def enable_recorder(self) -> None:
        if self._recorder_enabled:
            return
        self._recorder_enabled = True

        self._postprocessor.enable()
        await self._recorder.start()

    @aio_task_with_room_id
    async def disable_recorder(self, force: bool = False) -> None:
        if not self._recorder_enabled:
            return
        self._recorder_enabled = False

        if force:
            self._postprocessor.disable()
            await self._recorder.stop()
        else:
            await self._recorder.stop()
            await self._postprocessor.wait()
            self._postprocessor.disable()

    async def update_info(self) -> None:
        await self._live.update_info()

    @aio_task_with_room_id
    async def update_session(self) -> None:
        if self._monitor_enabled:
            await self._danmaku_client.stop()

        await self._live.deinit()
        await self._live.init()
        self._danmaku_client.session = self._live.session

        if self._monitor_enabled:
            await self._danmaku_client.start()

    async def _setup(self) -> None:
        self._setup_danmaku_client()
        self._setup_live_monitor()
        self._setup_live_event_submitter()
        self._setup_recorder()
        self._setup_postprocessor()
        self._setup_postprocessor_event_submitter()

    def _setup_danmaku_client(self) -> None:
        self._danmaku_client = DanmakuClient(
            self._live.session, self._live.room_id
        )

    def _setup_live_monitor(self) -> None:
        self._live_monitor = LiveMonitor(self._danmaku_client, self._live)

    def _setup_live_event_submitter(self) -> None:
        self._live_event_submitter = LiveEventSubmitter(self._live_monitor)

    def _setup_recorder(self) -> None:
        self._recorder = Recorder(
            self._live,
            self._danmaku_client,
            self._live_monitor,
            self._out_dir,
            self._path_template,
            buffer_size=self._buffer_size,
            read_timeout=self._read_timeout,
            danmu_uname=self._danmu_uname,
            filesize_limit=self._filesize_limit,
            duration_limit=self._duration_limit,
        )

    def _setup_postprocessor_event_submitter(self) -> None:
        self._postprocessor_event_submitter = \
            PostprocessorEventSubmitter(self._postprocessor)

    def _setup_postprocessor(self) -> None:
        self._postprocessor = Postprocessor(
            self._live,
            self._recorder,
            remux_to_mp4=self._remux_to_mp4,
            delete_source=self._delete_source,
        )

    async def _destroy(self) -> None:
        self._destroy_postprocessor_event_submitter()
        self._destroy_postprocessor()
        self._destroy_recorder()
        self._destroy_live_event_submitter()
        self._destroy_live_monitor()
        self._destroy_danmaku_client()

    def _destroy_danmaku_client(self) -> None:
        del self._danmaku_client

    def _destroy_live_monitor(self) -> None:
        del self._live_monitor

    def _destroy_live_event_submitter(self) -> None:
        del self._live_event_submitter

    def _destroy_recorder(self) -> None:
        del self._recorder

    def _destroy_postprocessor_event_submitter(self) -> None:
        del self._postprocessor_event_submitter

    def _destroy_postprocessor(self) -> None:
        del self._postprocessor



from __future__ import annotations
import asyncio
from typing import Dict, Iterator, TYPE_CHECKING

from blrec.setting.models import OutputSettings

from .task import RecordTask
from .models import TaskData, TaskParam, FileDetail
from ..exception import NotFoundError
if TYPE_CHECKING:
    from ..setting import SettingsManager
from ..setting import (
    HeaderSettings,
    DanmakuSettings,
    RecorderSettings,
    PostprocessingSettings,
    TaskSettings,
)


__all__ = 'RecordTaskManager',


class RecordTaskManager:
    def __init__(self, settings_manager: SettingsManager) -> None:
        self._settings_manager = settings_manager
        self._tasks: Dict[int, RecordTask] = {}
        import threading
        self._lock = threading.Lock()

    async def load_all_tasks(self) -> None:
        settings_list = self._settings_manager.get_settings({'tasks'}).tasks
        assert settings_list is not None

        for settings in settings_list:
            await self.add_task(settings)

    async def destroy_all_tasks(self) -> None:
        if not self._tasks:
            return
        await asyncio.wait([t.destroy() for t in self._tasks.values()])
        self._tasks.clear()

    def has_task(self, room_id: int) -> bool:
        return room_id in self._tasks

    async def add_task(self, settings: TaskSettings) -> None:
        task = RecordTask(settings.room_id)
        self._tasks[settings.room_id] = task

        await self._settings_manager.apply_task_header_settings(
            settings.room_id, settings.header, update_session=False
        )
        await task.setup()

        self._settings_manager.apply_task_output_settings(
            settings.room_id, settings.output
        )
        self._settings_manager.apply_task_danmaku_settings(
            settings.room_id, settings.danmaku
        )
        self._settings_manager.apply_task_recorder_settings(
            settings.room_id, settings.recorder
        )
        self._settings_manager.apply_task_postprocessing_settings(
            settings.room_id, settings.postprocessing
        )

        if settings.enable_monitor:
            await task.enable_monitor()
        if settings.enable_recorder:
            await task.enable_recorder()

    async def remove_task(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.disable_recorder(force=True)
        await task.disable_monitor()
        await task.destroy()
        del self._tasks[room_id]

    async def remove_all_tasks(self) -> None:
        coros = [self.remove_task(i) for i in self._tasks]
        if coros:
            await asyncio.wait(coros)

    async def start_task(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.update_info()
        await task.enable_monitor()
        await task.enable_recorder()

    async def stop_task(self, room_id: int, force: bool = False) -> None:
        task = self._get_task(room_id)
        await task.disable_recorder(force)
        await task.disable_monitor()

    async def start_all_tasks(self) -> None:
        await self.update_all_task_infos()
        await self.enable_all_task_monitors()
        await self.enable_all_task_recorders()

    async def stop_all_tasks(self, force: bool = False) -> None:
        await self.disable_all_task_recorders(force)
        await self.disable_all_task_monitors()

    async def enable_task_monitor(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.enable_monitor()

    async def disable_task_monitor(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.disable_monitor()

    async def enable_all_task_monitors(self) -> None:
        coros = [t.enable_monitor() for t in self._tasks.values()]
        if coros:
            await asyncio.wait(coros)

    async def disable_all_task_monitors(self) -> None:
        coros = [t.disable_monitor() for t in self._tasks.values()]
        if coros:
            await asyncio.wait(coros)

    async def enable_task_recorder(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.enable_recorder()

    async def disable_task_recorder(
            self, room_id: int, force: bool = False
    ) -> None:
        task = self._get_task(room_id)
        await task.disable_recorder(force)

    async def enable_all_task_recorders(self) -> None:
        coros = [t.enable_recorder() for t in self._tasks.values()]
        if coros:
            await asyncio.wait(coros)

    async def disable_all_task_recorders(self, force: bool = False) -> None:
        coros = [t.disable_recorder(force) for t in self._tasks.values()]
        if coros:
            await asyncio.wait(coros)

    def get_task_data(self, room_id: int) -> TaskData:
        task = self._get_task(room_id)
        assert task.ready, "the task isn't ready yet, couldn't get task data!"
        return self._make_task_data(task)

    def get_all_task_data(self) -> Iterator[TaskData]:
        for task in filter(lambda t: t.ready, self._tasks.values()):
            yield self._make_task_data(task)

    def get_task_param(self, room_id: int) -> TaskParam:
        task = self._get_task(room_id)
        return self._make_task_param(task)

    def get_task_file_details(self, room_id: int) -> Iterator[FileDetail]:
        task = self._get_task(room_id)
        for path in task.files:
            yield FileDetail.from_path(path)

    async def update_task_info(self, room_id: int) -> None:
        task = self._get_task(room_id)
        await task.update_info()

    async def update_all_task_infos(self) -> None:
        coros = [t.update_info() for t in self._tasks.values()]
        if coros:
            await asyncio.wait(coros)

    async def apply_task_header_settings(
        self,
        room_id: int,
        settings: HeaderSettings,
        *,
        update_session: bool = True,
    ) -> None:
        task = self._get_task(room_id)

        # avoid unnecessary updates that will interrupt connections
        if (
            task.user_agent == settings.user_agent and
            task.cookie == settings.cookie
        ):
            return

        task.user_agent = settings.user_agent
        task.cookie = settings.cookie

        if update_session:
            # update task session to take the effect
            await task.update_session()

    def apply_task_output_settings(
        self, room_id: int, settings: OutputSettings
    ) -> None:
        task = self._get_task(room_id)
        task.out_dir = settings.out_dir
        task.path_template = settings.path_template
        task.filesize_limit = settings.filesize_limit
        task.duration_limit = settings.duration_limit

    def apply_task_danmaku_settings(
        self, room_id: int, settings: DanmakuSettings
    ) -> None:
        task = self._get_task(room_id)
        task.danmu_uname = settings.danmu_uname

    def apply_task_recorder_settings(
        self, room_id: int, settings: RecorderSettings
    ) -> None:
        task = self._get_task(room_id)
        task.quality_number = settings.quality_number
        task.read_timeout = settings.read_timeout
        task.buffer_size = settings.buffer_size

    def apply_task_postprocessing_settings(
        self, room_id: int, settings: PostprocessingSettings
    ) -> None:
        task = self._get_task(room_id)
        task.remux_to_mp4 = settings.remux_to_mp4
        task.delete_source = settings.delete_source

    def _get_task(self, room_id: int) -> RecordTask:
        try:
            return self._tasks[room_id]
        except KeyError:
            raise NotFoundError(f'no task for the room {room_id}')

    def _make_task_param(self, task: RecordTask) -> TaskParam:
        return TaskParam(
            out_dir=task.out_dir,
            path_template=task.path_template,
            filesize_limit=task.filesize_limit,
            duration_limit=task.duration_limit,
            user_agent=task.user_agent,
            cookie=task.cookie,
            danmu_uname=task.danmu_uname,
            quality_number=task.quality_number,
            read_timeout=task.read_timeout,
            buffer_size=task.buffer_size,
            remux_to_mp4=task.remux_to_mp4,
            delete_source=task.delete_source,
        )

    def _make_task_data(self, task: RecordTask) -> TaskData:
        return TaskData(
            user_info=task.user_info,
            room_info=task.room_info,
            task_status=task.status,
        )



from enum import IntEnum
from datetime import datetime
from typing import cast

import attr

from .typing import ResponseData


__all__ = 'LiveStatus', 'RoomInfo', 'UserInfo'


class LiveStatus(IntEnum):
    PREPARING = 0
    LIVE = 1
    ROUND = 2


@attr.s(auto_attribs=True, frozen=True, slots=True)
class RoomInfo:
    uid: int
    room_id: int
    short_room_id: int
    area_id: int
    area_name: str
    parent_area_id: int
    parent_area_name: str
    live_status: LiveStatus
    live_start_time: int  # a integer in seconds
    online: int
    title: str
    cover: str
    tags: str
    description: str

    @staticmethod
    def from_data(data: ResponseData) -> 'RoomInfo':
        if (timestamp := data.get('live_start_time', None)) is not None:
            live_start_time = cast(int, timestamp)
        elif (time_string := data.get('live_time', None)) is not None:
            if time_string == '0000-00-00 00:00:00':
                live_start_time = 0
            else:
                dt = datetime.fromisoformat(time_string)
                live_start_time = int(dt.timestamp())
        else:
            raise ValueError(f'Failed to init live_start_time: {data}')

        return RoomInfo(
            uid=data['uid'],
            room_id=int(data['room_id']),
            short_room_id=int(data['short_id']),
            area_id=data['area_id'],
            area_name=data['area_name'],
            parent_area_id=data['parent_area_id'],
            parent_area_name=data['parent_area_name'],
            live_status=LiveStatus(data['live_status']),
            live_start_time=live_start_time,
            online=int(data['online']),
            title=data['title'],
            cover=data.get('cover', None) or data.get('user_cover', None),
            tags=data['tags'],
            description=data['description'],
        )


@attr.s(auto_attribs=True, frozen=True, slots=True)
class UserInfo:
    name: str
    gender: str
    face: str
    uid: int
    level: int
    sign: str

    @staticmethod
    def from_data(data: ResponseData) -> 'UserInfo':
        return UserInfo(
            name=data['name'],
            gender=data['sex'],
            face=data['face'],
            uid=data['mid'],
            level=data['level'],
            sign=data['sign'],
        )
